# -*- coding: utf-8 -*-
"""
:Module:         khoros.utils.tests.test_mentions
:Synopsis:       This module is used by pytest to verify that user and content mentions work properly
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  11 May 2020
"""

import os
import re
import sys

import pytest

from khoros.objects import messages
from khoros.errors import exceptions
from khoros import Khoros

CORRECT_CONTENT_MENTION = '<li-message title="Click Here" uid="6560" url=' + \
                          '"https://community.khoros.com/t5/Community-FAQs/' + \
                          'Understanding-SEO-Friendly-URLs/ta-p/6560"></li-message>'
CORRECT_USER_MENTION = '<li-user uid="1" login="@admin"></li-user>'


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionadded:: 2.4.0
    """
    sys.path.insert(0, os.path.abspath('../..'))
    return


def expected_content_response(response):
    """This function replaces the TLD to match the example constant and then identifies if the response was expected.

    .. versionadded:: 2.4.0

    :param response: The response received from the function
    :type response: str
    :returns: Boolean value indicating if the response was expected
    """
    response = re.sub(r'https://.*\.com/t5', 'https://community.khoros.com/t5', response)
    return True if response == CORRECT_CONTENT_MENTION else False


def expected_user_response(response):
    """This function identifies if the response was expected.

    .. versionadded:: 2.4.0

    :param response: The response received from the function
    :type response: str
    :returns: Boolean value indicating if the response was expected
    """
    return True if response == CORRECT_USER_MENTION else False


def get_content_test_data(include_id=True, false_id=False, relative_url=False):
    """This function returns the test data to use in the various test functions.

    .. versionadded:: 2.4.0

    :param include_id: Determines if the Content ID should be returned (``True`` by default)
    :type include_id: bool
    :param false_id: Determines if an incorrect Content ID should be returned (``False`` by default)
    :type false_id: bool
    :param relative_url: Determines if a relative URL should be returned (``False`` by default)
    :returns: A tuple containing the appropriate test data
    """
    title = "Click Here"
    url = "https://community.khoros.com/t5/Community-FAQs/Understanding-SEO-Friendly-URLs/ta-p/6560"
    content_id = "6560"
    false_content_id = "12345"
    if relative_url:
        url = url.split('khoros.com')[1]
    data = [title, url]
    if include_id:
        if false_id:
            data.append(false_content_id)
        else:
            data.append(content_id)
    return tuple(data)


def get_user_test_data():
    """This function returns the test data to use in the user mention tests.

    .. versionadded:: 2.4.0

    :returns: The User ID and login as a tuple of two strings
    """
    user_id = '1'
    login = 'admin'
    return user_id, login


def test_content_mention_with_all_arguments():
    """This function tests the :py:func:`khoros.objects.messages.format_content_mention` when all required arguments
       have been supplied.

    .. versionadded:: 2.4.0

    :returns: None
    """
    title, url, content_id = get_content_test_data()
    response = messages.format_content_mention(content_id=content_id, title=title, url=url)
    assert response == CORRECT_CONTENT_MENTION
    return


def test_content_mention_with_full_dict():
    """This function tests the :py:func:`khoros.objects.messages.format_content_mention` when all required dictionary
       keys and values have been supplied.

    .. versionadded:: 2.4.0

    :returns: None
    """
    title, url, content_id = get_content_test_data()
    content_info = {
        'id': content_id,
        'title': title,
        'url': url
    }
    response = messages.format_content_mention(content_info=content_info)
    assert response == CORRECT_CONTENT_MENTION
    return


@pytest.mark.filterwarnings("error")
def test_content_mention_with_no_id_arg():
    """This function tests creating a content mention when no Content ID has been supplied as an argument.

    .. versionadded:: 2.4.0

    :returns: None
    """
    title, url = get_content_test_data(include_id=False)
    response = messages.format_content_mention(title=title, url=url)
    assert response == CORRECT_CONTENT_MENTION
    return


@pytest.mark.filterwarnings("error")
def test_content_mention_with_no_id_dict():
    """This function tests creating a content mention when no Content ID has been supplied in the dictionary.

    .. versionadded:: 2.4.0

    :returns: None
    """
    title, url = get_content_test_data(include_id=False)
    content_info = {'title': title, 'url': url}
    response = messages.format_content_mention(content_info=content_info)
    assert response == CORRECT_CONTENT_MENTION
    return


def test_content_mention_with_false_id_arg():
    """This function tests creating a content mention when an invalid Content ID is supplied as an argument.

    .. versionadded:: 2.4.0

    :returns: None
    """
    title, url, content_id = get_content_test_data(false_id=True)
    with pytest.warns(UserWarning):
        response = messages.format_content_mention(content_id=content_id, title=title, url=url)
    assert response == CORRECT_CONTENT_MENTION
    return


def test_content_mention_with_false_id_dict():
    """This function tests creating a content mention when an invalid Content ID is supplied in the dictionary.

    .. versionadded:: 2.4.0

    :returns: None
    """
    title, url, content_id = get_content_test_data(false_id=True)
    content_info = {
        'id': content_id,
        'title': title,
        'url': url
    }
    with pytest.warns(UserWarning):
        response = messages.format_content_mention(content_info=content_info)
    assert response == CORRECT_CONTENT_MENTION
    return


def test_bad_content_url():
    """This function tests creating a content mention when an invalid URL is supplied.

    .. versionadded:: 2.4.0

    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.MessageTypeNotFoundError`
    """
    title = "Click Here"
    url = "https://khoros.com/platform/communities"

    # Testing with arguments
    with pytest.raises(exceptions.MessageTypeNotFoundError):
        messages.format_content_mention(title=title, url=url)

    # Testing with content dictionary
    content_info = {'title': title, 'url': url}
    with pytest.raises(exceptions.MessageTypeNotFoundError):
        messages.format_content_mention(content_info=content_info)
    return


def test_relative_content_url_without_object():
    """This function tests creating a content mention with a relative content URL and no Khoros object.

    .. versionadded:: 2.4.0

    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    title, url, content_id = get_content_test_data(relative_url=True)

    # Testing arguments with and without a Content ID
    with pytest.raises(exceptions.MissingRequiredDataError):
        messages.format_content_mention(title=title, url=url)
        messages.format_content_mention(content_id=content_id, title=title, url=url)

    # Testing dictionary with and without a Content ID
    content_info = {'title': title, 'url': url}
    with pytest.raises(exceptions.MissingRequiredDataError):
        messages.format_content_mention(content_info=content_info)
        content_info['id'] = content_id
        messages.format_content_mention(content_info=content_info)
    return


@pytest.mark.skip(reason="Session key doesn't work in GitHub Actions CI.")
def test_relative_content_url_with_object():
    """This function tests creating a content mention with a relative content URL but with a Khoros object.

    .. versionadded:: 2.4.0

    :returns: None
    """
    title, url, content_id = get_content_test_data(relative_url=True)
    khoros = Khoros()
    print()

    # Testing arguments without a Content ID
    response = messages.format_content_mention(khoros, title=title, url=url)
    assert expected_content_response(response)

    # Testing arguments with a Content ID
    response = messages.format_content_mention(khoros, content_id=content_id, title=title, url=url)
    assert expected_content_response(response)

    # Testing dictionary without a Content ID
    content_info = {'title': title, 'url': url}
    response = messages.format_content_mention(khoros, content_info)
    assert expected_content_response(response)

    # Testing dictionary with a Content ID
    content_info['id'] = content_id
    response = messages.format_content_mention(khoros, content_info)
    assert expected_content_response(response)
    return


def test_user_mention_with_arguments():
    """This function tests creating a user mention with all required arguments provided.

    .. versionadded:: 2.4.0

    :returns: None
    """
    user_id, login = get_user_test_data()
    response = messages.format_user_mention(user_id=user_id, login=login)
    assert expected_user_response(response)
    return


def test_user_mention_with_dictionary():
    """This function tests creating a user mention with all required dictionary key value pairs provided.

    .. versionadded:: 2.4.0

    :returns: None
    """
    user_id, login = get_user_test_data()
    user_info = {'id': user_id, 'login': login}
    response = messages.format_user_mention(user_info=user_info)
    assert expected_user_response(response)
    return


def test_user_mention_with_args_no_object():
    """This function tests creating a user mention with a single argument and no Khoros object.

    .. versionadded:: 2.4.0

    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.MissingAuthDataError`
    """
    user_id, login = get_user_test_data()

    # Test with no login
    with pytest.raises(exceptions.MissingAuthDataError):
        messages.format_user_mention(user_id=user_id)

    # Test with no User ID
    with pytest.raises(exceptions.MissingAuthDataError):
        messages.format_user_mention(login=login)
    return


def test_user_mention_with_dictionary_no_object():
    """This function tests creating a user mention with a single key value pair and no Khoros object.

    .. versionadded:: 2.4.0

    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.MissingAuthDataError`
    """
    user_id, login = get_user_test_data()
    user_info_id = {'id': user_id}
    user_info_login = {'login': login}

    # Test with no login
    with pytest.raises(exceptions.MissingAuthDataError):
        messages.format_user_mention(user_info=user_info_id)

    # Test with no User ID
    with pytest.raises(exceptions.MissingAuthDataError):
        messages.format_user_mention(user_info=user_info_login)
    return


@pytest.mark.skip(reason="Session key doesn't work in GitHub Actions CI.")
def test_user_mention_with_dict_and_object():
    """This function tests creating a user mention with one key value pair and with a Khoros object.

    .. versionadded:: 2.4.0

    :returns: None
    """
    user_id, login = get_user_test_data()
    user_info_id = {'id': user_id}
    user_info_login = {'login': login}
    khoros = Khoros()

    # Test with no login
    response = messages.format_user_mention(khoros, user_info_id)
    assert expected_user_response(response)

    # Test with no User ID
    response = messages.format_user_mention(khoros, user_info_login)
    assert expected_user_response(response)
    return


# TODO: Add test for not existent user



# -*- coding: utf-8 -*-
"""
:Module:            khoros.structures.communities
:Synopsis:          This module contains functions specific to the high-level community configuration
:Usage:             ``from khoros.structures import communities``
:Example:           ``details = get_community_details(khoros_object)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     26 Apr 2020
"""

from . import base
from .. import errors
from ..utils.core_utils import display_warning


def get_community_details(khoros_object):
    """This function returns a dictionary of community configuration settings.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :returns: The community details within a dictionary
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    response = base.get_details(khoros_object, community=True)
    _check_for_multiple_tenants(response)
    return response['data']['items'][0]


def _check_for_multiple_tenants(_community_details):
    """This function checks to see if more than one community instance (i.e. tenant) was found and displays a warning.

    .. versionadded:: 2.1.0

    :param _community_details: Dictionary containing community details from LiQL
    :type _community_details: dict
    :returns: None
    """
    if _community_details['data']['size'] > 1:
        display_warning(f"{_community_details['data']['size']} community instances (i.e. tenants) were found but " +
                        "only details from the first tenant will be returned.")
    return


def get_community_field(khoros_object, field, community_details=None):
    """This function returns a specific community field from the Khoros Community API.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param field: The field from the :py:class:`khoros.structures.base.Mapping` class whose value should be returned
    :type field: str
    :param community_details: The data captured from the :py:func:`khoros.structures.base.get_details` function
    :type community_details: dict, None
    :returns: The requested field in its native format
    :raises: :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.InvalidStructureTypeError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    return base.get_structure_field(khoros_object, field, community=True, details=community_details)


def get_tenant_id(khoros_object, community_details=None):
    """This function retrieves the tenant ID of the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The tenant ID in string format
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'id', community_details)


def get_title(khoros_object, full_title=True, short_title=False, community_details=None):
    """This function retrieves the full and/or short title of the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param full_title: Return the full title of the environment (``True`` by default)
    :type full_title: bool
    :param short_title: Return the short title of the environment (``False`` by default)
    :type short_title: bool
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The title(s) of the environment as a string or a tuple of strings
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    if not full_title and not short_title:
        exc_msg = "Must return at least the full title or the short title."
        errors.exceptions.MissingRequiredDataError(exc_msg)
    if not community_details:
        community_details = get_community_details(khoros_object)
    titles = (community_details['title'], community_details['short_title'])
    if not short_title:
        titles = titles[0]
    elif not full_title:
        titles = titles[1]
    return titles


def get_description(khoros_object, community_details=None):
    """This function retrieves the description of the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The description in string format
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'description', community_details)


def get_primary_url(khoros_object, community_details=None):
    """This function retrieves the primary URL of the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The primary URL in string format
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'view_href', community_details)


def get_max_attachments(khoros_object, community_details=None):
    """This function retrieves the maximum number of attachments permitted per message within the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The value as an integer
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'attachment_max_per_message', community_details)


def get_permitted_attachment_types(khoros_object, community_details=None):
    """This function retrieves the attachment file types permitted within the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The permitted file types within a list
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'attachment_file_types', community_details).split(',')


def email_confirmation_required_to_post(khoros_object, community_details=None):
    """This function identifies if an email configuration is required before posting in the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: Boolean value indicating if email configuration is required before posting
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'email_confirmation_required_to_post', community_details)


def get_language(khoros_object, community_details=None):
    """This function retrieves the language (e.g. ``en``) utilized in the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The language code as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'language', community_details)


def get_ooyala_player_branding_id(khoros_object, community_details=None):
    """This function retrieves the branding ID for the Ooyala Player utilized within the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The branding ID in string format
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    if not community_details:
        community_details = get_community_details(khoros_object)
    if 'ooyala_player_branding_id' not in community_details:
        branding_id = None
    else:
        branding_id = community_details['ooyala_player_branding_id']
    return branding_id


def get_date_pattern(khoros_object, community_details=None):
    """This function retrieves the date pattern (e.g. ``yyyy-MM-dd``) utilized within the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The date pattern in string format
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'date_pattern', community_details)


def friendly_date_enabled(khoros_object, community_details=None):
    """This function if the friendly date functionality is utilized within the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: Boolean value indicating if the feature is enabled
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'friendly_date_enabled', community_details)


def get_friendly_date_max_age(khoros_object, community_details=None):
    """This function identifies if the friendly date functionality is utilized within the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: Boolean value indicating if the feature is enabled
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'friendly_date_max_age', community_details)


def get_active_skin(khoros_object, community_details=None):
    """This function retrieves the primary active skin that is utilized within the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The skin name as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'skin', community_details)


def get_sign_out_url(khoros_object, community_details=None):
    """This function retrieves the Sign Out URL for the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The Sign Out URL as a string
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'web_ui_sign_out_url', community_details)


def get_creation_date(khoros_object, community_details=None):
    """This function retrieves the timestamp for the initial creation of the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: The creation date as a string (e.g. ``2020-02-03T22:41:36.408-08:00``)
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    # TODO: Allow a format to be specified and the ability to parse as a datetime object if needed
    return get_community_field(khoros_object, 'creation_date', community_details)


def top_level_categories_enabled(khoros_object, community_details=None):
    """This function identifies if top level categories are enabled within the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: Boolean value indicating if top level categories are enabled
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'top_level_categories_enabled', community_details)


def show_community_node_in_breadcrumb(khoros_object, community_details=None):
    """This function identifies if the community node should be shown in breadcrumbs.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: Boolean value indicating if the community node is displayed in bredcrumbs
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'tlc_show_community_node_in_breadcrumb', community_details)


def show_breadcrumb_at_top_level(khoros_object, community_details=None):
    """This function identifies if breadcrumbs should be shown at the top level of the environment.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: Boolean value indicating if breadcrumbs are displayed at the top level of the environment
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'tlc_show_breadcrumb_at_top_level', community_details)


def top_level_categories_on_community_page(khoros_object, community_details=None):
    """This function identifies if top level categories are enabled on community pages.

    .. versionadded:: 2.1.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param community_details: Dictionary containing community details (optional)
    :type community_details: dict, None
    :returns: Boolean value indicating if top level categories are enabled on community pages
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    return get_community_field(khoros_object, 'tlc_set_on_community_page', community_details)



# -*- coding: utf-8 -*-
"""
:Module:            khoros.api
:Synopsis:          This module handles interactions with the Khoros Community REST APIs
:Usage:             ``import khoros.api``
:Example:           ``json_response = khoros.api.get_request_with_retries(url, auth_dict=khoros.auth)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     16 May 2020
"""

import json

import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder

from . import errors
from .utils import core_utils


def define_headers(khoros_object=None, auth_dict=None, params=None, accept=None, content_type=None, multipart=False):
    """This function defines the headers to use in an API call.

    .. versionchanged:: 2.3.0
       Added the ``multipart`` Boolean argument to remove the ``Content-Type`` key value pair when appropriate.

    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param params: Header parameters in a dictionary format
    :type params: dict, None
    :param accept: The ``Accept`` header value (e.g. ``application/json``)
    :type accept: str, None
    :param content_type: The ``Content-Type`` header value (e.g. ``application/json``)
    :type content_type: str, None
    :param multipart: Defines whether or not the query is a ``multipart/form-data`` query (``False`` by default)
    :type multipart: bool
    :returns: A dictionary with the header fields and associated values
    """
    if not khoros_object and not auth_dict:
        raise errors.exceptions.MissingAuthDataError()
    else:
        if auth_dict:
            headers = auth_dict['header']
        else:
            headers = khoros_object.auth['header']
    if params:
        headers.update(params)
    if accept:
        headers['Accept'] = accept
    if content_type:
        headers['Content-Type'] = content_type
    if multipart and 'Content-Type' in headers:
        del headers['Content-Type']
    return headers


def _get_json_query_string(_return_json, _include_ampersand_prefix=True):
    """This function constructs a query string for a Community API v1 query that should return JSON responses.

    :param _return_json: Indicates whether or not API responses should be in JSON format
    :type _return_json: bool
    :param _include_ampersand_prefix: Determines if an ampersand (``&``) prefix should be included in the string
    :type _include_ampersand_prefix: bool
    :returns: The formatted query string
    """
    _query_strings = {True: 'restapi.response_format=json', False: ''}
    _prefixes = {True: '&', False: ''}
    _query_string = ''
    if _return_json:
        _query_string = f"{_prefixes.get(_include_ampersand_prefix)}{_query_strings.get(_return_json)}"
    return _query_string


def get_request_with_retries(query_url, return_json=True, khoros_object=None, auth_dict=None, headers=None):
    """This function performs a GET request with a total of 5 retries in case of timeouts or connection issues.

    .. versionchanged:: 2.5.0
       Leverages new private functions and has improved error handling of JSON conversion attempts.

    :param query_url: The URI to be queried
    :type query_url: str
    :param return_json: Determines whether or not the response should be returned in JSON format (Default: ``True``)
    :type return_json: bool
    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param headers: Any header values (in dictionary format) to pass in the API call (optional)
    :type headers: dict, None
    :returns: The API response from the GET request (optionally in JSON format)
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`
    """
    headers = define_headers(khoros_object=khoros_object, auth_dict=auth_dict, params=headers)
    retries = 0
    while retries <= 5:
        try:
            response = requests.get(query_url, headers=headers)
            break
        except Exception as exc_msg:
            _report_failed_attempt(exc_msg, 'get', retries)
            retries += 1
            pass
    if retries == 6:
        _raise_exception_for_repeated_timeouts()
    return _attempt_json_conversion(response, return_json)


def _api_request_with_payload(_url, _payload=None, _request_type='post', _headers=None, _multipart=False):
    """This function performs an API request while supplying a JSON payload.

    .. versionchanged:: 2.5.0
       The function can now be called without supplying a JSON payload.

    .. versionchanged:: 2.3.0
       Added the ability to perform multipart/form-data queries.

    :param _url: The URI to be queried
    :type _url: str
    :param _payload: The payload that accompanies the API call
    :type _payload: dict, str, None
    :param _request_type: Define the type of API call (e.g. ``get``, ``post`` or ``put``)
    :type _request_type: str
    :param _headers: Any predefined headers to be passed with the API call (optional)
    :type _headers: dict
    :param _multipart: Defines whether or not the query is a ``multipart/form-data`` query (``False`` by default)
    :type _multipart: bool
    :returns: The API response
    :raises: :py:exc:`khoros.errors.exceptions.InvalidRequestTypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`
    """
    _headers = {} if not _headers else _headers
    _retries = 0
    if not _payload:
        _response = _api_request_without_payload(_url, _request_type, _headers)
    else:
        while _retries <= 5:
            try:
                if _request_type.lower() == "put":
                    if _multipart:
                        _response = requests.put(_url, files=_payload, headers=_headers)
                    else:
                        _response = requests.put(_url, data=json.dumps(_payload, default=str), headers=_headers)
                elif _request_type.lower() == "post":
                    if _multipart:
                        _response = requests.post(_url, files=_payload, headers=_headers)
                    else:
                        _response = requests.post(_url, data=json.dumps(_payload, default=str), headers=_headers)
                else:
                    raise errors.exceptions.InvalidRequestTypeError()
                break
            except Exception as _exc_msg:
                _report_failed_attempt(_exc_msg, _request_type, _retries)
                _retries += 1
                pass
        if _retries == 6:
            _raise_exception_for_repeated_timeouts()
    return _response


def _api_request_without_payload(_url, _request_type, _headers):
    """This function performs a ``POST`` or ``PUT`` request without an accompanying JSON payload.

    :param _url: The URL for the API request
    :type _url: str
    :param _request_type: The request type (e.g. ``post`` or ``put``)
    :type _request_type: str
    :param _headers: The headers associated with the API request
    :type _headers: dict
    :returns: The API response
    :raises: :py:exc:`khoros.errors.exceptions.InvalidRequestTypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`
    """
    _retries = 0
    while _retries <= 5:
        try:
            if _request_type.lower() == "post":
                _response = requests.post(_url, headers=_headers)
            elif _request_type.lower() == "put":
                _response = requests.put(_url, headers=_headers)
            else:
                raise errors.exceptions.InvalidRequestTypeError()
            break
        except Exception as _exc_msg:
            _report_failed_attempt(_exc_msg, _request_type, _retries)
            _retries += 1
            pass
    if _retries == 6:
        _raise_exception_for_repeated_timeouts()
    return _response


def _report_failed_attempt(_exc_msg, _request_type, _retries):
    """This function reports a failed API call that will be retried.

    :param _exc_msg: The exception that was raised can captured within a try/except clause
    :param _request_type: The type of API request (e.g. ``post``, ``put`` or ``get``)
    :type _request_type: str
    :param _retries: The attempt number for the API request
    :type _retries: int
    :returns: None
    """
    _exc_name = type(_exc_msg).__name__
    if 'connect' not in _exc_name.lower():
        raise Exception(f"{_exc_name}: {_exc_msg}")
    _current_attempt = f"(Attempt {_retries} of 5)"
    _error_msg = f"The {_request_type.upper()} request has failed with the following exception: " + \
                 f"{_exc_name}: {_exc_msg} {_current_attempt}"
    errors.handlers.eprint(f"{_error_msg}\n{_exc_name}: {_exc_msg}\n")
    return


def _raise_exception_for_repeated_timeouts():
    """This function raises an exception when all API attempts (including) retries resulted in a timeout.

    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.APIConnectionError`
    """
    _failure_msg = "The script was unable to complete successfully after five consecutive API timeouts. " + \
                   "Please run the script again or contact Khoros Support for further assistance."
    raise errors.exceptions.APIConnectionError(_failure_msg)


def post_request_with_retries(url, json_payload=None, return_json=True, khoros_object=None, auth_dict=None,
                              headers=None, multipart=False):
    """This function performs a POST request with a total of 5 retries in case of timeouts or connection issues.

    .. versionchanged:: 2.5.0
       The function can now be called without supplying a JSON payload.

    .. versionchanged:: 2.3.0
       Added the ability to perform multipart/form-data queries.

    :param url: The URI to be queried
    :type url: str
    :param json_payload: The payload for the POST request in JSON format
    :type json_payload: dict, None
    :param return_json: Determines whether or not the response should be returned in JSON format (Default: ``True``)
    :type return_json: bool
    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param headers: Any header values (in dictionary format) to pass in the API call (optional)
    :type headers: dict, None
    :param multipart: Defines whether or not the query is a ``multipart/form-data`` query (``False`` by default)
    :type multipart: bool
    :returns: The API response from the POST request
    :raises: :py:exc:`ValueError`, :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`
    """
    headers = define_headers(khoros_object=khoros_object, auth_dict=auth_dict, params=headers, multipart=multipart)
    response = _api_request_with_payload(url, json_payload, 'post', headers, multipart)
    return _attempt_json_conversion(response, return_json)


def put_request_with_retries(url, json_payload=None, return_json=True, khoros_object=None, auth_dict=None,
                             headers=None, multipart=False):
    """This function performs a PUT request with a total of 5 retries in case of timeouts or connection issues.

    .. versionchanged:: 2.5.0
       The function can now be called without supplying a JSON payload.

    .. versionchanged:: 2.3.0
       Added the ability to perform multipart/form-data queries.

    :param url: The URI to be queried
    :type url: str
    :param json_payload: The payload for the PUT request in JSON format
    :type json_payload: dict, None
    :param return_json: Determines whether or not the response should be returned in JSON format (Default: ``True``)
    :type return_json: bool
    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param headers: Any header values (in dictionary format) to pass in the API call (optional)
    :type headers: dict, None
    :param multipart: Defines whether or not the query is a ``multipart/form-data`` query (``False`` by default)
    :type multipart: bool
    :returns: The API response from the PUT request
    :raises: :py:exc:`ValueError`, :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.PUTRequestError`
    """
    headers = define_headers(khoros_object=khoros_object, auth_dict=auth_dict, params=headers, multipart=multipart)
    response = _api_request_with_payload(url, json_payload, 'put', headers, multipart)
    return _attempt_json_conversion(response, return_json)


def _attempt_json_conversion(_response, _return_json):
    """This function attempts to convert an API response to JSON if requested.

    .. versionadded:: 2.5.0

    :param _response: The API response to be converted
    :param _return_json: Indicates whether or not the API response should be converted
    :type _return_json: bool
    :returns: The API response that has been converted to JSON or in its original format if unable to convert
    """
    if _return_json and not isinstance(_response, dict):
        try:
            _response = _response.json()
        except Exception as _exc_msg:
            _exc_name = type(_exc_msg).__name__
            errors.handlers.eprint(f"Failed to convert to JSON due to the following exception: {_exc_name}: {_exc_msg}")
    return _response


def query_successful(api_response):
    """This function reviews the API response from the Community API to verify whether or not the call was successful.

    :param api_response: The response from the API in JSON format
    :type api_response: dict
    :returns: Boolean indicating whether or not the API call was successful
    """
    try:
        success_values = ['successful', 'success']
        successful = True if api_response['status'] in success_values else False
    except (KeyError, IndexError, ValueError, TypeError):
        successful = False
    return successful


def get_results_count(api_response):
    """This function returns the number of results within a response from the Community API.

    :param api_response: The response to an API query in JSON format
    :type api_response: dict
    :returns: The number of results in the API response as an integer
    """
    return api_response['data']['size']


def get_items_list(api_response):
    """This function returns the list of ``items`` dictionaries within a response from the Community API.

    :param api_response: The response to an API query in JSON format
    :type api_response: dict
    :returns: List of ``items`` dictionaries from the API response
    """
    return api_response['data']['items']


def delete(url, return_json=False, khoros_object=None, auth_dict=None, headers=None):
    """This function performs a DELETE request against the Core API.

    :param url: The URI against which the DELETE request will be issued
    :type url: str
    :param return_json: Determines whether or not the response should be returned in JSON format (Default: ``False``)
    :type return_json: bool
    :param khoros_object: The core Khoros object (Required if the ``auth_dict`` parameter is not supplied)
    :type khoros_object: class[khoros.Khoros], None
    :param auth_dict: The ``auth`` dictionary within the :py:class:`khoros.Khoros` class object
    :type auth_dict: dict, None
    :param headers: Any header values (in dictionary format) to pass in the API call (optional)
    :type headers: dict, None
    :returns: The API response from the DELETE request (optionally in JSON format)
    """
    headers = define_headers(khoros_object=khoros_object, auth_dict=auth_dict, params=headers)
    response = requests.delete(url, headers=headers)
    if return_json:
        response = response.json()
    return response


def perform_v1_search(khoros_object, endpoint, filter_field, filter_value, return_json=False, fail_on_no_results=False):
    """This function performs a search for a particular field value using a Community API v1 call.

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param endpoint: The API v1 endpoint against which to perform the search query
    :type endpoint: str
    :param filter_field: The name of the field being queried within the API v1 endpoint
    :type filter_field: str
    :param filter_value: The value associated with the field being queried
    :type filter_value: str, int
    :param return_json: Determines if the response should be returned in JSON format (``False`` by default)
    :type return_json: bool
    :param fail_on_no_results: Raises an exception if no results are returned (``False`` by default)
    :type fail_on_no_results: bool
    :returns: The API response (optionally in JSON format)
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    headers = define_headers(khoros_object, content_type='application/x-www-form-urlencoded')
    if type(filter_value) == str:
        filter_value = core_utils.url_encode(filter_value)
    uri = f"{khoros_object.core['v1_base']}/search/{endpoint}?q={filter_field}:{filter_value}"
    uri = f"{uri}{_get_json_query_string(return_json)}"
    response = requests.get(uri, headers=headers)
    if return_json:
        response = response.json()
        response = response['response'] if 'response' in response else response
    errors.handlers.verify_v1_response(response, 'get', 'users', fail_on_no_results)
    return response


def encode_multipart_data(data_fields):
    """This function uses the Streaming Multipart Data Encoder to encode the payload for a multipart/form-data API call.

    .. versionadded:: 2.3.0

    .. seealso:: This function follows the `requests_toolbelt <https://rsa.im/3dg7QiZ>`_ documentation.

    :param data_fields: A dictionary with the data to be encoded
    :type data_fields: dict
    :returns: The encoded data for use by the :py:mod:`requests` library
    """
    return MultipartEncoder(fields=data_fields)


def encode_v1_query_string(query_dict, return_json=True):
    """This function formats and URL-encodes a Community API v1 quey string.

    .. versionadded:: 2.5.0

    :param query_dict: A dictionary with the query fields and associated values
    :type query_dict: dict
    :param return_json: Determines if JSON should be returned rather than XML (default: ``True``)
    :type return_json: bool
    :returns: The properly formatted and encoded query string
    """
    if return_json:
        query_dict['restapi.response_format'] = 'json'
    return core_utils.encode_query_string(query_dict)


def make_v1_request(khoros_object, endpoint, query_params, request_type='GET', return_json=True):
    """This function makes a Community API v1 request.

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param endpoint: The API endpoint to be queried
    :type endpoint: str
    :param query_params: The field and associated values to be leveraged in the query string
    :type query_params: dict
    :param request_type: Determines which type of API request to perform (e.g. ``GET``, ``POST``, ``PUT``, etc.)
    :type request_type: str
    :param return_json: Determines if the response should be returned in JSON format rather than the default
    :type return_json: bool
    :returns: The API response
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`,
             :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`,
             :py:exc:`khoros.errors.exceptions.PUTRequestError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.CurrentlyUnsupportedError`,
             :py:exc:`khoros.errors.exceptions.InvalidRequestTypeError`
    """
    currently_unsupported_types = ['UPDATE', 'PATCH', 'DELETE']
    query_string = encode_v1_query_string(query_params, return_json)
    header = {"Content-Type": "application/x-www-form-urlencoded"}
    url = f"{khoros_object.core['v1_base']}/{endpoint}?{query_string}"
    print(url)
    if request_type.upper() == 'GET':
        response = get_request_with_retries(url, return_json, khoros_object, headers=header)
    elif request_type.upper() == 'POST':
        response = post_request_with_retries(url, return_json=return_json, khoros_object=khoros_object, headers=header)
    elif request_type.upper() == 'PUT':
        response = put_request_with_retries(url, return_json=return_json, khoros_object=khoros_object, headers=header)
    elif request_type.upper() in currently_unsupported_types:
        errors.exceptions.CurrentlyUnsupportedError()
    else:
        raise errors.exceptions.InvalidRequestTypeError()
    # TODO: Verify successful response
    return response


def deliver_v2_results(response, full_response=None, return_id=None, return_url=None, return_api_url=None,
                       return_http_code=None, return_status=None, return_developer_message=None):
    """This function parses a Community API v2 response and returned specific data based on the function arguments.

    .. versionadded:: 2.5.0
       The code for this function was extracted from the :py:func:`khoros.objects.messages.create` function.

    :param response: The API response to be parsed
    :param full_response: Determines if the full raw API response should be returned
    :type full_response: bool, None
    :param return_id: Determines if the ``id`` field value should be returned
    :type return_id: bool, None
    :param return_url: Determines if the ``view_href`` field value should be returned
    :type return_url: bool, None
    :param return_api_url: Determines if the ``href`` field value should be returned
    :type return_api_url: bool, None
    :param return_http_code: Determines if the ``http_code`` field value should be returned
    :type return_http_code: bool, None
    :param return_status: Determines if the ``http_code`` field value should be returned
    :type return_status: bool, None
    :param return_developer_message: Determines if the ``http_code`` field value should be returned
    :returns: Boolean value indicating a successful outcome (default), the full API response or one or more specific
              fields defined by function arguments
    """
    outcome = query_successful(response)
    if any((return_id, return_url, return_api_url, return_http_code, return_status, return_developer_message)):
        return_values = {
            'return_id': parse_v2_response(response, data_id=True),
            'return_url': parse_v2_response(response, data_url=True),
            'return_api_url': parse_v2_response(response, data_api_uri=True),
            'return_http_code': parse_v2_response(response, http_code=True),
            'return_status': parse_v2_response(response, status=True),
            'return_developer_message': parse_v2_response(response, developer_msg=True),
        }
        data_to_return = []
        return_booleans = {
            'return_id': return_id,
            'return_url': return_url,
            'return_http_code': return_http_code,
            'return_api_url': return_api_url,
            'return_status': return_status,
            'return_developer_message': return_developer_message,
        }
        for return_key, return_value in return_booleans.items():
            if return_value:
                data_to_return.append(return_values.get(return_key))
        outcome = tuple(data_to_return)
        if len(data_to_return) == 1:
            outcome = outcome[0]
    return response if full_response else outcome


def parse_v2_response(json_response, return_dict=False, status=False, developer_msg=False, http_code=False,
                      data_id=False, data_url=False, data_api_uri=False, v2_base=''):
    """This function parses an API response for a Community API v2 operation and returns parsed data.

    .. versionchanged:: 2.5.0
       Moved from the :py:mod:`khoros.objects.messages` module to :py:mod:`khoros.api` and expanded the scope to
       apply to most Community API v2 API responses.

    .. versionadded:: 2.3.0

    :param json_response: The API response in JSON format
    :type json_response: dict
    :param return_dict: Defines if the parsed data should be returned within a dictionary
    :type return_dict: bool
    :param status: Defines if the **status** value should be returned
    :type status: bool
    :param developer_msg: Defines if the **developer response** message should be returned
    :type developer_msg: bool
    :param http_code: Defines if the **HTTP status code** should be returned
    :type http_code: bool
    :param data_id: Defines if the **ID** should be returned
    :type data_id: bool
    :param data_url: Defines if the **URL** should be returned
    :type data_url: bool
    :param data_api_uri: Defines if the **API URI** should be returned
    :type data_api_uri: bool
    :param v2_base: The base URL for the API v2
    :type v2_base: str
    :returns: A string, tuple or dictionary with the parsed data
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    parsed_data = {}
    fields = {
        'status': (status, ('status',)),
        'developer_msg': (developer_msg, ('message',)),
        'http_code': (http_code, ('http_code',)),
        'data_id': (data_id, ('data', 'id')),
        'data_url': (data_url, ('data', 'view_href')),
        'data_api_uri': (data_api_uri, ('data', 'href')),
    }
    _confirm_field_supplied(fields)
    for field, info in fields.items():
        requested, json_path = info[0], info[1]
        if requested:
            if len(json_path) == 1:
                value = json_response[json_path[0]]
            else:
                value = json_response[json_path[0]][json_path[1]]
            parsed_data[field] = value
    if 'data_api_uri' in parsed_data and v2_base != '':
        parsed_data['data_api_uri'] = f"{v2_base}/{parsed_data.get('data_api_uri')}"
    if not return_dict:
        parsed_data = tuple(list(parsed_data.values()))
        if len(parsed_data) == 1:
            parsed_data = parsed_data[0]
    return parsed_data


def _confirm_field_supplied(_fields_dict):
    """This function checks to ensure that at least one field has been enabled to retrieve.

    .. versionchanged:: 2.5.0
       Moved from the :py:mod:`khoros.objects.messages` module to :py:mod:`khoros.api`.

    .. versionadded:: 2.3.0
    """
    _field_supplied = False
    for _field_value in _fields_dict.values():
        if _field_value[0]:
            _field_supplied = True
            break
    if not _field_supplied:
        raise errors.exceptions.MissingRequiredDataError("At least one field must be enabled to retrieve a response.")
    return



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.helper
:Synopsis:          Module that allows the khoros library to leverage a helper configuration file
:Usage:             ``from khoros.utils import helper``
:Example:           ``helper_settings = helper.get_settings('/tmp/helper.yml', 'yaml')``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     26 Apr 2020
"""

import json

import yaml

from .. import errors
from .core_utils import get_file_type


# Define function to import a YAML helper file
def import_helper_file(file_path, file_type):
    """This function imports a YAML (.yml) or JSON (.json) helper config file.

    .. versionchanged:: 2.2.0
       Changed the name and replaced the ``yaml.load`` function call with ``yaml.safe_load`` to be more secure.

    :param file_path: The file path to the YAML file
    :type file_path: str
    :param file_type: Defines the file type as either ``yaml`` or ``json``
    :type file_type: str
    :returns: The parsed configuration data
    :raises: :py:exc:`FileNotFoundError`, :py:exc:`khoros.errors.exceptions.InvalidHelperFileTypeError`
    """
    with open(file_path, 'r') as cfg_file:
        if file_type == 'yaml':
            helper_cfg = yaml.safe_load(cfg_file)
        elif file_type == 'json':
            helper_cfg = json.load(cfg_file)
        else:
            raise errors.exceptions.InvalidHelperFileTypeError
    return helper_cfg


# Define function to covert a YAML Boolean value to a Python Boolean value
def _convert_yaml_to_bool(_yaml_bool_value):
    """This function converts the 'yes' and 'no' YAML values to traditional Boolean values."""
    true_values = ['yes', 'true']
    if _yaml_bool_value.lower() in true_values:
        _bool_value = True
    else:
        _bool_value = False
    return _bool_value


# Define function to get the connection information
def _get_connection_info(_helper_cfg):
    """This function parses any connection information found in the helper file.

    .. versionchanged:: 2.2.0
       Removed one of the preceding underscores in the function name
    """
    _connection_info = {}
    _connection_keys = ['community_url', 'tenant_id', 'default_auth_type']
    for _key in _connection_keys:
        if _key in _helper_cfg['connection']:
            _connection_info[_key] = _helper_cfg['connection'][_key]

    # Parse OAuth 2.0 information if found
    if 'oauth2' in _helper_cfg['connection']:
        _connection_info['oauth2'] = _get_oauth2_info(_helper_cfg)

    # Parse session authentication information if found
    if 'session_auth' in _helper_cfg['connection']:
        _connection_info['session_auth'] = _get_session_auth_info(_helper_cfg)
    return _connection_info


def _get_oauth2_info(_helper_cfg):
    """This function parses OAuth 2.0 information if found in the helper file.

    .. versionchanged:: 2.2.0
       Removed one of the preceding underscores in the function name
    """
    _oauth2 = {}
    _oauth2_keys = ['client_id', 'client_secret', 'redirect_url']
    for _key in _oauth2_keys:
        if _key in _helper_cfg['connection']['oauth2']:
            _oauth2[_key] = _helper_cfg['connection']['oauth2'][_key]
        else:
            _oauth2[_key] = ''
    return _oauth2


def _get_session_auth_info(_helper_cfg):
    """This function parses session authentication information if found in the helper file.

    .. versionchanged:: 2.2.0
       Removed one of the preceding underscores in the function name
    """
    _session_auth = {}
    _session_info = ['username', 'password']
    for _key in _session_info:
        if _key in _helper_cfg['connection']['session_auth']:
            _session_auth[_key] = _helper_cfg['connection']['session_auth'][_key]
        else:
            _session_auth[_key] = None
    return _session_auth


def _get_construct_info(_helper_cfg):
    """This function parses settings that can be leveraged in constructing API responses and similar tasks.

    .. versionchanged:: 2.2.0
       Removed one of the preceding underscores in the function name
    """
    _construct_info = {}
    _top_level_keys = ['prefer_json']
    for _key in _top_level_keys:
        if _key in _helper_cfg:
            _key_val = _helper_cfg[_key]
            if _key_val in HelperParsing.yaml_boolean_values:
                _key_val = HelperParsing.yaml_boolean_values.get(_key_val)
            _construct_info[_key] = _key_val
        else:
            _construct_info[_key] = None
    return _construct_info


# Define function to retrieve the helper configuration settings
def get_helper_settings(file_path, file_type='yaml'):
    """This function returns a dictionary of the defined helper settings.

    .. versionchanged:: 2.2.0
       Added support for JSON formatted helper configuration files

    :param file_path: The file path to the helper configuration file
    :type file_path: str
    :param file_type: Defines the helper configuration file as a ``yaml`` file (default) or a ``json`` file
    :type file_type: str
    :returns: Dictionary of helper variables
    :raises: :py:exc:`khoros.errors.exceptions.InvalidHelperFileTypeError`
    """
    # Initialize the helper_settings dictionary
    helper_settings = {}

    if file_type != 'yaml' and file_type != 'json':
        file_type = get_file_type(file_path)

    # Import the helper configuration file
    helper_cfg = import_helper_file(file_path, file_type)

    # Populate the connection information in the helper dictionary
    if 'connection' in helper_cfg:
        helper_settings['connection'] = _get_connection_info(helper_cfg)

    # Populate the construct information in the helper dictionary
    helper_settings['construct'] = _get_construct_info(helper_cfg)

    # Return the helper_settings dictionary
    return helper_settings


# Define class for dictionaries to help in parsing the configuration files
class HelperParsing:
    """This class is used to help parse values imported from a YAML configuration file."""
    # Define dictionary to map YAML Boolean to Python Boolean
    yaml_boolean_values = {
        True: True,
        False: False,
        'yes': True,
        'no': False
    }



# -*- coding: utf-8 -*-
"""
:Module:            khoros.errors.exceptions
:Synopsis:          Collection of exception classes relating to the khoros library
:Usage:             ``import khoros.errors.exceptions``
:Example:           ``raise khoros.errors.exceptions.BadCredentialsError``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     11 May 2020
"""

#################
# Base Exception
#################


# Define base exception class
class KhorosError(Exception):
    """This is the base class for Khoros exceptions."""
    pass


#########################
# Base Object Exceptions
#########################


class InvalidNodeTypeError(KhorosError):
    """This exception is used when an invalid node type is provided."""
    def __init__(self, *args, **kwargs):
        default_msg = "The node type that was provided is invalid."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('node type ')[0]}'{kwargs['val']}'{default_msg.split('The')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class NodeIDNotFoundError(KhorosError):
    """This exception is used when a valid Node ID could not be found in a provided URL."""
    def __init__(self, *args, **kwargs):
        default_msg = "A valid Node ID could not be identified in the given URL."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('URL')[0]}: {kwargs['val']}"
            args = (custom_msg,)
        super().__init__(*args)


class NodeTypeNotFoundError(KhorosError):
    """This exception is used when a valid node type could not be found in a provided URL."""
    def __init__(self, *args, **kwargs):
        default_msg = "A valid node type could not be identified in the given URL."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('URL')[0]}: {kwargs['val']}"
            args = (custom_msg,)
        super().__init__(*args)


############################
# Base Structure Exceptions
############################


class InvalidStructureTypeError(KhorosError):
    """This exception is used when an invalid node type is provided.

    .. versionadded:: 2.1.0
    """
    def __init__(self, *args, **kwargs):
        default_msg = "The structure type that was provided is invalid."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('structure type ')[0]}'{kwargs['val']}'{default_msg.split('The')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


############################
# Authentication Exceptions
############################


class InvalidCallbackURLError(KhorosError):
    """This exception is used when an invalid Callback URL for OAuth 2.0 was not provided."""
    def __init__(self, *args, **kwargs):
        default_msg = "The OAuth 2.0 callback URL that was provided is invalid. The entire URL must be provided."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('URL ')[0]}'{kwargs['val']}'{default_msg.split('URL')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class MissingAuthDataError(KhorosError):
    """This exception is used when authentication data is not supplied and therefore a connection cannot occur."""
    def __init__(self, *args, **kwargs):
        default_msg = "The authentication data was not provided and a connection cannot be established."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class SessionAuthenticationError(KhorosError):
    """This exception is used when the session key authentication attempt failed."""
    def __init__(self, *args, **kwargs):
        default_msg = "The session key authentication attempt failed."
        if not (args or kwargs):
            args = (default_msg,)
        # TODO: Add alternate message if variable is passed to the exception class
        super().__init__(*args)


#####################
# General Exceptions
#####################


class CurrentlyUnsupportedError(KhorosError):
    """This exception is used when a feature or functionality being used is currently unsupported.

    .. versionchanged:: 2.0.0
       The unsupported feature can be passed as a string argument to explicitly reference it in the exception.
    """
    def __init__(self, *args, **kwargs):
        default_msg = "This feature is currently unsupported at this time."
        if not (args or kwargs):
            args = (default_msg,)
        else:
            custom_msg = f"The '{args[0]}' {default_msg.split('This ')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class DataMismatchError(KhorosError):
    """This exception is used when there is a mismatch between two data sources.

    .. versionadded:: 2.3.0
    """
    def __init__(self, *args, **kwargs):
        default_msg = "A data mismatch was found with the data sources."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'data' in kwargs:
            multi_types = [list, tuple, set]
            if type(kwargs['data']) == str:
                custom_msg = f"{default_msg.split('data')[0]}'{kwargs['val']}'{default_msg.split('with the')[1]}"
                custom_msg = custom_msg.replace('sources', 'source')
                args = (custom_msg,)
            elif type(kwargs['data']) in multi_types and len(kwargs['data']) == 2:
                custom_section = f"'{kwargs['data'][0]}' and '{kwargs['data'][1]}'"
                custom_msg = f"{default_msg.split('data sources')[0]}{custom_section}{default_msg.split('with the')[1]}"
                args = (custom_msg,)
        super().__init__(*args)


class InvalidFieldError(KhorosError):
    """This exception is used when an invalid field is provided.

    .. versionadded:: 2.1.0
    """
    def __init__(self, *args, **kwargs):
        default_msg = "The field that was provided is invalid."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'val' in kwargs:
            custom_msg = f"{default_msg.split('field ')[0]}'{kwargs['val']}'{default_msg.split('The')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class InvalidURLError(KhorosError):
    """This exception is used when a provided URL is invalid.

    .. versionadded:: 2.1.0
    """
    def __init__(self, *args, **kwargs):
        default_msg = "The provided URL is invalid"
        if not (args or kwargs):
            args = (default_msg,)
        elif 'url' in kwargs:
            custom_msg = f"{default_msg.split('is')[0]}'{kwargs['url']}'{default_msg.split('URL')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class MissingRequiredDataError(KhorosError):
    """This exception is used when a function or method is missing one or more required arguments.

    .. versionadded:: 2.0.0
    """
    def __init__(self, *args, **kwargs):
        default_msg = "Missing one or more required parameters"
        init_msg = "The object failed to initialize as it is missing one or more required arguments."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'init' in args or 'initialize' in args:
            if 'object' in kwargs:
                custom_msg = f"{init_msg.split('object')[0]}'{kwargs['object']}'{init_msg.split('The')[1]}"
                args = (custom_msg,)
            else:
                args = (init_msg,)
        super().__init__(*args)


class UnknownFileTypeError(KhorosError):
    """This exception is used when a file type of a given file cannot be identified.

    .. versionadded:: 2.2.0
    """
    def __init__(self, *args, **kwargs):
        default_msg = "The file type of the given file path cannot be identified."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'file' in kwargs:
            custom_msg = f"{default_msg.split('path')[0]}'{kwargs['file']}'{default_msg.split('path')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


#########################
# Generic API Exceptions
#########################


class APIConnectionError(KhorosError):
    """This exception is used when the API query could not be completed due to connection aborts and/or timeouts."""
    def __init__(self, *args, **kwargs):
        default_msg = "The API query could not be completed due to connection aborts and/or timeouts."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class APIRequestError(KhorosError):
    """This exception is used for generic API request errors when there isn't a more specific exception."""
    def __init__(self, *args, **kwargs):
        default_msg = "The DELETE request did not return a successful response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class DELETERequestError(KhorosError):
    """This exception is used for generic DELETE request errors when there isn't a more specific exception."""
    def __init__(self, *args, **kwargs):
        default_msg = "The DELETE request did not return a successful response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class GETRequestError(KhorosError):
    """This exception is used for generic GET request errors when there isn't a more specific exception."""
    def __init__(self, *args, **kwargs):
        default_msg = "The GET request did not return a successful response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class InvalidEndpointError(KhorosError):
    """This exception is used when an invalid API endpoint / service is provided."""
    def __init__(self, *args, **kwargs):
        default_msg = "The supplied endpoint for the API is not recognized. (Examples of valid " + \
                      "lookup types include 'people' and 'contents')"
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class InvalidLookupTypeError(KhorosError):
    """This exception is used when an invalid API lookup type is provided."""

    def __init__(self, *args, **kwargs):
        default_msg = "The supplied lookup type for the API is not recognized. (Examples of valid " + \
                      "lookup types include 'id' and 'email')"
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class InvalidRequestTypeError(KhorosError):
    """This exception is used when an invalid API request type is provided."""

    def __init__(self, *args, **kwargs):
        default_msg = "The supplied request type for the API is not recognized. (Examples of valid " + \
                      "request types include 'POST' and 'PUT')"
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class LookupMismatchError(KhorosError):
    """This exception is used when an a lookup value doesn't match the supplied lookup type."""
    def __init__(self, *args, **kwargs):
        default_msg = "The supplied lookup type for the API does not match the value that was provided."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class NotFoundResponseError(KhorosError):
    """This exception is used when an API query returns a 404 response and there isn't a more specific class."""
    def __init__(self, *args, **kwargs):
        default_msg = "The API query returned a 404 response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class POSTRequestError(KhorosError):
    """This exception is used for generic POST request errors when there isn't a more specific exception."""
    def __init__(self, *args, **kwargs):
        default_msg = "The POST request did not return a successful response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class PUTRequestError(KhorosError):
    """This exception is used for generic PUT request errors when there isn't a more specific exception."""
    def __init__(self, *args, **kwargs):
        default_msg = "The PUT request did not return a successful response."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


####################
# Helper Exceptions
####################


class InvalidHelperFileTypeError(KhorosError, ValueError):
    """This exception is used when an invalid file type is provided for the helper file."""
    def __init__(self, *args, **kwargs):
        default_msg = "The helper configuration file can only have the 'yaml' or 'json' file type."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class InvalidHelperArgumentsError(KhorosError):
    """This exception is used when the helper function was supplied arguments instead of keyword arguments."""
    def __init__(self, *args, **kwargs):
        default_msg = "The helper configuration file only accepts basic keyword arguments. (e.g. arg_name='arg_value')"
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class HelperFunctionNotFoundError(KhorosError):
    """This exception is used when a function referenced in the helper config file does not exist."""
    def __init__(self, *args, **kwargs):
        default_msg = "The function referenced in the helper configuration file could not be found."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


##################
# LiQL Exceptions
##################


class InvalidOperatorError(KhorosError):
    """This exception is used when an invalid operator is provided for the LiQL query."""
    def __init__(self, *args, **kwargs):
        default_msg = "An invalid operator was provided for the LiQL query."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class OperatorMismatchError(KhorosError):
    """This exception is used when the number of operators in the LiQL query does not match the number of fields."""
    def __init__(self, *args, **kwargs):
        default_msg = "The number of operators provided in the LiQL query does not match the number of fields/values."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


class TooManyResultsError(KhorosError):
    """This exception is used when more results are returned than were expected in a LiQL query.

    .. versionadded:: 2.0.0
    """
    def __init__(self, *args, **kwargs):
        default_msg = "The number of operators provided in the LiQL query does not match the number of fields/values."
        if not (args or kwargs):
            args = (default_msg,)
        super().__init__(*args)


#####################
# Message Exceptions
#####################


class MessageTypeNotFoundError(KhorosError):
    """This exception is used when a message type cannot be identified from a given URL."""

    def __init__(self, *args, **kwargs):
        default_msg = "The message type could not be identified in the provided URL."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'msg_type' in kwargs:
            custom_msg = f"{default_msg.split('message type ')[0]}'{kwargs['msg_type']}'{default_msg.split('type')[1]}"
            args = (custom_msg,)
        elif 'url' in kwargs:
            custom_msg = f"{default_msg.split('provided')[0]}following URL: {kwargs['url']}"
            args = (custom_msg,)
        super().__init__(*args)


##################
# Role Exceptions
##################


class InvalidRoleError(KhorosError):
    """This exception is used when an invalid role is provided."""
    def __init__(self, *args, **kwargs):
        default_msg = "The role is invalid"
        if not (args or kwargs):
            args = (default_msg,)
        elif 'role' in kwargs:
            custom_msg = f"{default_msg.split('role ')[0]}'{kwargs['role']}'{default_msg.split('role')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


class InvalidRoleTypeError(KhorosError):
    """This exception is used when an invalid role type is provided."""
    def __init__(self, *args, **kwargs):
        default_msg = "The role type is invalid"
        if not (args or kwargs):
            args = (default_msg,)
        elif 'role_type' in kwargs:
            custom_msg = f"{default_msg.split('type ')[0]}'{kwargs['role_type']}'{default_msg.split('type ')[1]}"
            args = (custom_msg,)
        super().__init__(*args)


##################
# User Exceptions
##################


class UserCreationError(KhorosError):
    """This exception is used when an attempt to create a user fails."""
    def __init__(self, *args, **kwargs):
        default_msg = "The user failed to be created."
        if not (args or kwargs):
            args = (default_msg,)
        elif 'user' in kwargs:
            custom_msg = f"{default_msg.split('user ')[0]}'{kwargs['user']}'{default_msg.split('user')[1]}"
            args = (custom_msg,)
        if 'exc_msg' in kwargs:
            full_msg = f"{args[0].split('.')[0]} due to the following exception: {kwargs['exc_msg']}"
            args = (full_msg,)
        super().__init__(*args)



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.messages
:Synopsis:          This module includes functions that handle messages within a Khoros Community environment
:Usage:             ``from khoros.objects import messages``
:Example:           ``response = messages.create_message(khoros_obj, 'My First Message', 'Hello World',
                    node_id='support-tkb')``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     09 May 2020
"""

import warnings

from . import attachments, users
from .. import api, liql, errors
from ..structures import nodes

REQUIRED_FIELDS = ['board', 'subject']
CONTEXT_KEYS = ['id', 'url']
SEO_KEYS = ['title', 'description', 'canonical_url']
MESSAGE_SEO_URLS = {
    'Blog Article': 'ba-p',
    'Blog Comment': 'bc-p',
    'Contest Item': 'cns-p',
    'Idea': 'idi-p',
    'Message': 'm-p',
    'Question': 'qaq-p',
    'TKB Article': 'ta-p',
    'Topic': 'td-p'
}


def create(khoros_object, subject=None, body=None, node=None, node_id=None, node_url=None, canonical_url=None,
           context_id=None, context_url=None, cover_image=None, images=None, is_answer=None, is_draft=None,
           labels=None, product_category=None, products=None, read_only=None, seo_title=None, seo_description=None,
           tags=None, teaser=None, topic=None, videos=None, attachment_file_paths=None, full_response=False,
           return_id=False, return_url=False, return_api_url=False, return_http_code=False):
    """This function creates a new message within a given node.

    .. versionadded:: 2.3.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param subject: The title or subject of the message
    :type subject: str, None
    :param body: The body of the message in HTML format
    :type body: str, None
    :param node: A dictionary containing the ``id`` key and its associated value indicating the destination
    :type node: dict, None
    :param node_id: The ID of the node in which the message will be published
    :type node_id: str, None
    :param node_url: The URL of the node in which the message will be published

                     .. note:: This argument is necessary in the absence of the ``node`` and ``node_id`` arguments.

    :type node_url: str, None
    :param canonical_url: The search engine-friendly URL to the message
    :type canonical_url: str, None
    :param context_id: Metadata on a message to identify the message with an external identifier of your choosing
    :type context_id: str, None
    :param context_url: Metadata on a message representing a URL to associate with the message (external identifier)
    :type context_url: str, None
    :param cover_image: The cover image set for the message
    :type cover_image: dict, None
    :param images: The query to retrieve images uploaded to the message
    :type images: dict, None
    :param is_answer: Designates the message as an answer on a Q&A board
    :type is_answer: bool, None
    :param is_draft: Indicates whether or not the message is still a draft (i.e. unpublished)
    :type is_draft: bool, None
    :param labels: The query to retrieve labels applied to the message
    :type labels: dict, None
    :param product_category: The product category (i.e. container for ``products``) associated with the message
    :type product_category: dict, None
    :param products: The product in a product catalog associated with the message
    :type products: dict, None
    :param read_only: Indicates whether or not the message should be read-only or have replies/comments blocked
    :type read_only: bool, None
    :param seo_title: The title of the message used for SEO purposes
    :type seo_title: str, None
    :param seo_description: A description of the message used for SEO purposes
    :type seo_description: str, None
    :param tags: The query to retrieve tags applied to the message
    :type tags: dict, None
    :param teaser: The message teaser (used with blog articles)
    :type teaser: str, None
    :param topic: The root message of the conversation in which the message appears
    :type topic: dict, None
    :param videos: The query to retrieve videos uploaded to the message
    :type videos: dict, None
    :param attachment_file_paths: The full path(s) to one or more attachment (e.g. ``path/to/file1.pdf``)
    :type attachment_file_paths: str, tuple, list, set, None
    :param full_response: Defines if the full response should be returned instead of the outcome (``False`` by default)

                          .. caution:: This argument overwrites the ``return_id``, ``return_url``, ``return_api_url``
                                       and ``return_http_code`` arguments.

    :type full_response: bool
    :param return_id: Indicates that the **Message ID** should be returned (``False`` by default)
    :type return_id: bool
    :param return_url: Indicates that the **Message URL** should be returned (``False`` by default)
    :type return_url: bool
    :param return_api_url: Indicates that the **API URL** of the message should be returned (``False`` by default)
    :type return_api_url: bool
    :param return_http_code: Indicates that the **HTTP status code** of the response should be returned
                             (``False`` by default)
    :type return_http_code: bool
    :returns: Boolean value indicating a successful outcome (default) or the full API response
    :raises: :py:exc:`TypeError`, :py:exc:`ValueError`, :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.DataMismatchError`
    """
    api_url = f"{khoros_object.core['v2_base']}/messages"
    payload = construct_payload(subject, body, node, node_id, node_url, canonical_url, context_id, context_url,
                                is_answer, is_draft, read_only, seo_title, seo_description, teaser, tags, cover_image,
                                images, labels, product_category, products, topic, videos)
    multipart = True if attachment_file_paths else False
    if multipart:
        payload = attachments.construct_multipart_payload(payload, attachment_file_paths)
    response = api.post_request_with_retries(api_url, payload, khoros_object=khoros_object, multipart=multipart)
    return api.deliver_v2_results(response, full_response, return_id, return_url, return_api_url, return_http_code)


def construct_payload(subject=None, body=None, node=None, node_id=None, node_url=None, canonical_url=None,
                      context_id=None, context_url=None, is_answer=None, is_draft=None, read_only=None, seo_title=None,
                      seo_description=None, teaser=None, tags=None, cover_image=None, images=None, labels=None,
                      product_category=None, products=None, topic=None, videos=None):
    """This function constructs and properly formats the JSON payload for a messages API request.

    .. versionadded:: 2.3.0

    :param subject: The title or subject of the message
    :type subject: str, None
    :param body: The body of the message in HTML format
    :type body: str, None
    :param node: A dictionary containing the ``id`` key and its associated value indicating the destination
    :type node: dict, None
    :param node_id: The ID of the node in which the message will be published
    :type node_id: str, None
    :param node_url: The URL of the node in which the message will be published

                     .. note:: This argument is necessary in the absence of the ``node`` and ``node_id`` arguments.

    :type node_url: str, None
    :param canonical_url: The search engine-friendly URL to the message
    :type canonical_url: str, None
    :param context_id: Metadata on a message to identify the message with an external identifier of your choosing
    :type context_id: str, None
    :param context_url: Metadata on a message representing a URL to associate with the message (external identifier)
    :type context_url: str, None
    :param is_answer: Designates the message as an answer on a Q&A board
    :type is_answer: bool, None
    :param is_draft: Indicates whether or not the message is still a draft (i.e. unpublished)
    :type is_draft: bool, None
    :param read_only: Indicates whether or not the message should be read-only or have replies/comments blocked
    :type read_only: bool, None
    :param seo_title: The title of the message used for SEO purposes
    :type seo_title: str, None
    :param seo_description: A description of the message used for SEO purposes
    :type seo_description: str, None
    :param teaser: The message teaser (used with blog articles)
    :type teaser: str, None
    :param tags: The query to retrieve tags applied to the message
    :type tags: dict, None
    :param cover_image: The cover image set for the message
    :type cover_image: dict, None
    :param images: The query to retrieve images uploaded to the message
    :type images: dict, None
    :param labels: The query to retrieve labels applied to the message
    :type labels: dict, None
    :param product_category: The product category (i.e. container for ``products``) associated with the message
    :type product_category: dict, None
    :param products: The product in a product catalog associated with the message
    :type products: dict, None
    :param topic: The root message of the conversation in which the message appears
    :type topic: dict, None
    :param videos: The query to retrieve videos uploaded to the message
    :type videos: dict, None
    :returns: The properly formatted JSON payload
    """
    # Define the default payload structure
    payload = {
        "data": {
            "type": "message"
        }
    }

    # Ensure the required fields are defined
    _verify_required_fields(node, node_id, node_url, subject)

    # Define the destination
    if not node:
        if node_id:
            node = {"id": f"{node_id}"}
        else:
            node = {"id": f"{nodes.get_node_id(url=node_url)}"}
    payload['data']['board'] = node

    # Add supplied data where appropriate if string or Boolean
    supplied_data = {
        'body': (body, str),
        'subject': (subject, str),
        'canonical_url': (canonical_url, str),
        'context_id': (context_id, str),
        'context_url': (context_url, str),
        'is_answer': (is_answer, bool),
        'is_draft': (is_draft, bool),
        'read_only': (read_only, bool),
        'seo_title': (seo_title, str),
        'seo_description': (seo_description, str),
        'teaser': (teaser, str)
    }
    for field_name, field_value in supplied_data.items():
        if field_value[0]:
            if field_value[1] == str:
                payload['data'][field_name] = f"{field_value[0]}"
            elif field_value[1] == bool and type(field_value[0]) == str:
                # noinspection PyTypeChecker
                payload['data'][field_name] = bool(field_value[0])

    # TODO: Add functionality for non-string and non-Boolean arguments

    return payload


def _verify_required_fields(_node, _node_id, _node_url, _subject):
    """This function verifies that the required fields to create a message are satisfied.

    .. versionadded:: 2.3.0

    :param _node: A dictionary containing the ``id`` key and its associated value indicating the destination
    :type _node: dict
    :param _node_id: The ID of the node in which the message will be published
    :type _node_id: str
    :param _node_url: The URL of the node in which the message will be published

                     .. note:: This argument is necessary in the absence of the ``node`` and ``node_id`` arguments.

    :type _node_url: str
    :param _subject: The title or subject of the message
    :type _subject: str
    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    _requirements_satisfied = True
    if (not _node and not _node_id and not _node_url) or (_node is not None and type(_node) != dict) or not _subject:
        _requirements_satisfied = False
    elif _node and (not _node_id and not _node_url):
        _requirements_satisfied = False if 'id' not in _node else True
    if not _requirements_satisfied:
        raise errors.exceptions.MissingRequiredDataError("A node and subject must be defined when creating messages")
    return


def _confirm_field_supplied(_fields_dict):
    """This function checks to ensure that at least one field has been enabled to retrieve.

    .. versionadded:: 2.3.0
    """
    _field_supplied = False
    for _field_value in _fields_dict.values():
        if _field_value[0]:
            _field_supplied = True
            break
    if not _field_supplied:
        raise errors.exceptions.MissingRequiredDataError("At least one field must be enabled to retrieve a response.")
    return


def parse_v2_response(json_response, return_dict=False, status=False, response_msg=False, http_code=False,
                      message_id=False, message_url=False, message_api_uri=False, v2_base=''):
    """This function parses an API response for a message operation (e.g. creating a message) and returns parsed data.

    .. deprecated:: 2.5.0
       Use the :py:func:`khoros.api.parse_v2_response` function instead.

    .. versionadded:: 2.3.0

    :param json_response: The API response in JSON format
    :type json_response: dict
    :param return_dict: Defines if the parsed data should be returned within a dictionary
    :type return_dict: bool
    :param status: Defines if the **status** value should be returned
    :type status: bool
    :param response_msg: Defines if the **developer response** message should be returned
    :type response_msg: bool
    :param http_code: Defines if the **HTTP status code** should be returned
    :type http_code: bool
    :param message_id: Defines if the **message ID** should be returned
    :type message_id: bool
    :param message_url: Defines if the **message URL** should be returned
    :type message_url: bool
    :param message_api_uri: Defines if the ** message API URI** should be returned
    :type message_api_uri: bool
    :param v2_base: The base URL for the API v2
    :type v2_base: str
    :returns: A string, tuple or dictionary with the parsed data
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    warnings.warn(f"This function is deprecated and 'khoros.api.parse_v2_response' should be used.", DeprecationWarning)
    return api.parse_v2_response(json_response, return_dict, status, response_msg, http_code, message_id, message_url,
                                 message_api_uri, v2_base)


def get_id_from_url(url):
    """This function retrieves the message ID from a given URL.

    .. versionadded:: 2.4.0

    :param url: The URL from which the ID will be parsed
    :type url: str
    :returns: The ID associated with the message in string format
    :raises: :py:exc:`khoros.errors.exceptions.MessageTypeNotFoundError`
    """
    for msg_type in MESSAGE_SEO_URLS.values():
        if msg_type in url:
            return (url.split(f'{msg_type}/')[1]).split('#')[0]
    raise errors.exceptions.MessageTypeNotFoundError(url=url)


def _get_required_user_mention_data(_khoros_object, _user_info, _user_id, _login):
    _missing_data_error = "A User ID or login must be supplied to construct an user @mention"
    _info_fields = ['id', 'login']
    if not any((_user_info, _user_id, _login)):
        raise errors.exceptions.MissingRequiredDataError(_missing_data_error)
    elif not _user_id and not _login:
        if not any(_field in _info_fields for _field in _user_info):
            raise errors.exceptions.MissingRequiredDataError(_missing_data_error)
        else:
            if 'id' in _user_info:
                _user_id = _user_info.get('id')
            if 'login' in _user_info:
                _login = _user_info.get('login')
    if not _user_id or not _login:
        if not _khoros_object:
            raise errors.exceptions.MissingAuthDataError()
        if not _user_id:
            _user_id = users.get_user_id(_khoros_object, login=_login)
        elif not _login:
            _login = users.get_login(_khoros_object, user_id=_user_id)
    return _user_id, _login


def format_user_mention(khoros_object=None, user_info=None, user_id=None, login=None):
    """This function formats the ``<li-user>`` HTML tag for a user @mention.

    .. versionadded:: 2.4.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object

                          .. note:: This argument is necessary if only one of the user values (i.e. ``user_id`` or
                                    ``login``) are passed in the function, as a lookup will need to be performed to
                                    define the missing value.

    :type khoros_object: class[khoros.Khoros], None
    :param user_info: A dictionary containing the ``'id'`` and/or ``'login'`` key(s) with the user information

                      .. note:: This argument is necessary if the User ID and/or Login are not explicitly passed
                                using the ``user_id`` and/or ``login`` function arguments.

    :type user_info: dict, None
    :param user_id: The unique user identifier (i.e. User ID) for the user
    :type user_id: str, int, None
    :param login: The login (i.e. username) for the user
    :type login: str, None
    :returns: The properly formatted ``<li-user>`` HTML tag in string format
    :raises: :py:exc:`khoros.errors.exceptions.MissingAuthDataError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    user_id, login = _get_required_user_mention_data(khoros_object, user_info, user_id, login)
    mention_tag = f'<li-user uid="{user_id}" login="@{login}"></li-user>'
    return mention_tag


def _report_missing_id_and_retrieve(_content_id, _url):
    """This function displays a ``UserWarning`` message if needed and then retrieves the correct ID from the URL.

    .. versionadded:: 2.4.0

    :param _content_id: The missing or incorrect ID of the message
    :type _content_id: str, int, None
    :param _url: The full URL of the message
    :type _url: str
    :returns: The appropriate ID of the message where possible
    :raises: :py:exc:`khoros.errors.exceptions.MessageTypeNotFoundError`
    """
    if _content_id is not None:
        warnings.warn(f"The given ID '{_content_id}' is not found in the URL {_url} and will be verified.",
                      UserWarning)
    return get_id_from_url(_url)


def _check_for_bad_content_id(_content_id, _url):
    """This function confirms that a supplied Content ID is found within the provided URL.

    .. versionadded:: 2.4.0

    :param _content_id: The ID of the message
    :type _content_id: str, int, None
    :param _url: The full URL of the message
    :type _url: str
    :returns: The appropriate ID of the message where possible
    :raises: :py:exc:`khoros.errors.exceptions.MessageTypeNotFoundError`
    """
    if _content_id is None or str(_content_id) not in _url:
        _content_id = _report_missing_id_and_retrieve(_content_id, _url)
    return _content_id


def _get_required_content_mention_data(_khoros_object, _content_info, _content_id, _title, _url):
    _missing_data_error = "A title and URL must be supplied to construct an user @mention"
    _content_info = {} if _content_info is None else _content_info
    _info_fields = ['title', 'url']
    _info_arguments = (_content_info, _content_id, _title, _url)
    _required_fields_in_dict = all(_field in _content_info for _field in _info_fields)
    _required_fields_in_args = all((_title, _url))
    if not _required_fields_in_dict and not _required_fields_in_args:
        raise errors.exceptions.MissingRequiredDataError(_missing_data_error)
    elif _required_fields_in_dict:
        _title = _content_info.get('title')
        _url = _content_info.get('url')
        _content_id = _content_info.get('id') if 'id' in _content_info else _content_id
    else:
        _content_id = _content_info.get('id') if not _content_id else _content_id
    _content_id = _check_for_bad_content_id(_content_id, _url)
    return _content_id, _title, _url


def format_content_mention(khoros_object=None, content_info=None, content_id=None, title=None, url=None):
    """This function formats the ``<li-message>`` HTML tag for a content @mention.

    .. versionadded:: 2.4.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object

                          .. note:: This argument is necessary if the URL (i.e. ``url`` argument) is not an absolute
                                    URL, as the base community URL will need to be retrieved from the object.

    :type khoros_object: class[khoros.Khoros], None
    :param content_info: A dictionary containing the ``'id'`` and/or ``'login'`` key(s) with the user information

                         .. note:: This argument is necessary if the Title and URL are not explicitly passed
                                   using the ``title`` and ``url`` function arguments.

    :type content_info: dict, None
    :param content_id: The Message ID (aka Content ID) associated with the content mention

                       .. note:: This is an optional argument as the ID can be retrieved from the URL.

    :type content_id: str, int, None
    :param title: The display title for the content mention (e.g. ``"Click Here"``)
    :type title: str, None
    :param url: The fully-qualified URL of the message being mentioned
    :type url: str, None
    :returns: The properly formatted ``<li-message>`` HTML tag in string format
    :raises: :py:exc:`khoros.errors.exceptions.MessageTypeNotFoundError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.MessageTypeNotFoundError`,
             :py:exc:`khoros.errors.exceptions.InvalidURLError`
    """
    content_id, title, url = _get_required_content_mention_data(khoros_object, content_info, content_id, title, url)
    if url.startswith('/t5'):
        if not khoros_object:
            raise errors.exceptions.MissingRequiredDataError('The core Khoros object is required when a '
                                                             'fully-qualified URL is not provided.')
        url = f"{khoros_object.core['base_url']}{url}"
    mention_tag = f'<li-message title="{title}" uid="{content_id}" url="{url}"></li-message>'
    return mention_tag



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.attachments
:Synopsis:          This module includes functions that handle attachments for messages
:Usage:             ``from khoros.objects import attachments``
:Example:           ``payload = format_attachment_payload(titles, file_paths)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     07 May 2020
"""

import os
import json

from .. import errors
from ..utils import core_utils


def construct_multipart_payload(message_json, file_paths):
    """This function constructs the full multipart payload for a message with one or more attachment.

    .. versionadded:: 2.3.0

    :param message_json: The message information in JSON format
    :type message_json: dict
    :param file_paths: The full path(s) to one or more attachment (e.g. ``path/to/file1.pdf``)
    :type file_paths: str, tuple, list, set
    :returns: The full payload for the multipart API call as a dictionary

              .. note:: See the `Requests Documentation <https://rsa.im/2SvMsya>`_ for added context on the return data.

    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.DataMismatchError`
    """
    file_paths = core_utils.convert_string_to_tuple(file_paths)
    message_json['data']['attachments'] = format_attachment_payload(file_paths)
    files_payload = get_file_upload_info(file_paths)
    full_payload = {'api.request': (None, json.dumps(message_json, default=str), 'application/json')}
    full_payload.update(files_payload)
    return full_payload


def format_attachment_payload(file_paths):
    """This function formats the JSON payload for attachments to be used in an API call.

    .. versionadded:: 2.3.0

    :param file_paths: The full path(s) to one or more attachment (e.g. ``path/to/file1.pdf``)
    :type file_paths: str, tuple, list, set
    :returns: The list of items (in list format) that represent the ``items`` API value
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.DataMismatchError`
    """
    file_paths = core_utils.convert_string_to_tuple(file_paths)
    payload = {
        "list_item_type": "attachment"
    }
    list_items = get_list_items(file_paths)
    payload["items"] = list_items
    return payload


def get_list_items(file_paths):
    """This function constructs the ``items`` field for the ``attachments`` API call.

    .. versionadded:: 2.3.0

    :param file_paths: The full path(s) to one or more attachment (e.g. ``path/to/file1.pdf``)
    :type file_paths: str, tuple, list, set
    :returns: The list of items (in list format) that represent the ``items`` API value
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.DataMismatchError`
    """
    file_paths = core_utils.convert_string_to_tuple(file_paths)
    list_items, count = [], 1
    for path in file_paths:
        item = {
            "type": "attachment",
            "field": f"attachment{count}",
            "filename": f"{os.path.basename(path)}"
        }
        list_items.append(item)
        count += 1
    return list_items


def get_file_upload_info(file_paths):
    """This function constructs the binary file(s) portion of the multipart API call.

    .. versionadded:: 2.3.0

    :param file_paths: The full path(s) to one or more attachment (e.g. ``path/to/file1.pdf``)
    :type file_paths: str, tuple, list, set
    :returns: A dictionary with the file upload information for the API call
    """
    file_paths = core_utils.convert_string_to_tuple(file_paths)
    files, count = {}, 1
    for path in file_paths:
        # TODO: Dynamically define the MIME type
        files[f'attachment{count}'] = (f'{os.path.basename(path)}', open(path, 'rb'))
        count += 1
    return files



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects
:Synopsis:          This module contains sub-modules that are specific to the various Community API objects.
:Usage:             ``from khoros.objects import base``
:Example:           ``node_id = base.get_node_id('https://community.khoros.com/t5/Khoros-Blog/bg-p/relnote', 'blog')``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     11 May 2020
"""

__all__ = ['albums', 'messages', 'roles', 'users']

# Import all submodules by default
from . import albums
from . import messages
from . import roles
from . import users
